# -*- coding: utf-8 -*-

# FOGLAMP_BEGIN
# See: http://foglamp.readthedocs.io/
# FOGLAMP_END

from aiohttp import web

from foglamp.common.logger import FLCoreLogger
from foglamp.common.storage_client.payload_builder import PayloadBuilder
from foglamp.services.core import connect

__author__ = "Ashish Jabble"
__copyright__ = "Copyright (c) 2023, Dianomic Systems Inc."
__license__ = "Apache 2.0"
__version__ = "${VERSION}"

_help = """
    ----------------------------------------------------------------
    | GET  DELETE          | /foglamp/monitors                      |
    | GET  DELETE          | /foglamp/monitors/{service}            |
    | GET  DELETE          | /foglamp/monitors/{service}/{counter}  |
    ----------------------------------------------------------------
"""
_LOGGER = FLCoreLogger().get_logger(__name__)

def setup(app):
    app.router.add_route('GET', '/foglamp/monitors', get_all)
    app.router.add_route('GET', '/foglamp/monitors/{service}', get_by_service_name)
    app.router.add_route('GET', '/foglamp/monitors/{service}/{counter}', get_by_service_and_counter_name)
    app.router.add_route('DELETE', '/foglamp/monitors', purge_all)
    app.router.add_route('DELETE', '/foglamp/monitors/{service}', purge_by_service)
    app.router.add_route('DELETE', '/foglamp/monitors/{service}/{counter}', purge_by_service_and_counter)

# TODO: 8167 - Limit and Offset support and other pending queries

async def get_all(request: web.Request) -> web.Response:
    """ GET list of performance monitors

    :Example:
        curl -sX GET http://localhost:8081/foglamp/monitors
    """
    storage = connect.get_storage_async()
    monitors = await storage.query_tbl("monitors")
    counters = monitors["rows"]
    monitor = {}
    for c in counters:
        val = {"average": c["average"], "maximum": c["maximum"], "minimum": c["minimum"], "samples": c["samples"],
               "timestamp": c["ts"], "service": c["service"]}
        monitor.setdefault(c['monitor'], []).append(val)
    monitors = [{'monitor': k, 'values': v} for k, v in monitor.items()]
    # Group by service name
    grouped_data = {}
    for entry in monitors:
        monitor_name = entry['monitor']
        values = entry['values']
        for value in values:
            service = value.pop('service', None)
            if service:
                if service not in grouped_data:
                    grouped_data[service] = {}
                if monitor_name not in grouped_data[service]:
                    grouped_data[service][monitor_name] = []
                # Group by monitor
                grouped_data[service][monitor_name].append(value)
    return web.json_response({"monitors": grouped_data})


async def get_by_service_name(request: web.Request) -> web.Response:
    """ GET performance monitors for the given service

    :Example:
        curl -sX GET http://localhost:8081/foglamp/monitors/<SVC_NAME>
    """
    service = request.match_info.get('service', None)
    storage = connect.get_storage_async()
    payload = PayloadBuilder().SELECT("average", "maximum", "minimum", "monitor", "samples", "ts").ALIAS(
        "return", ("ts", 'timestamp')).FORMAT("return", ("ts", "YYYY-MM-DD HH24:MI:SS.MS")).WHERE(
        ["service", '=', service]).payload()
    response = {"service": service}
    result = await storage.query_tbl_with_payload('monitors', payload)
    if 'rows' in result:
        monitor = {}
        for row in result["rows"]:
            val = {"average": row["average"], "maximum": row["maximum"], "minimum": row["minimum"],
                   "samples": row["samples"], "timestamp": row["timestamp"]}
            monitor.setdefault(row['monitor'], []).append(val)
        monitors = [{'monitor': k, 'values': v} for k, v in monitor.items()]
        response["monitors"] = monitors
    return web.json_response(response)

async def get_by_service_and_counter_name(request: web.Request) -> web.Response:
    """ GET values for the single counter for the single service

    :Example:
        curl -sX GET http://localhost:8081/foglamp/monitors/<SVC_NAME>/<COUNTER_NAME>
    """
    service = request.match_info.get('service', None)
    counter = request.match_info.get('counter', None)

    storage = connect.get_storage_async()
    payload = PayloadBuilder().SELECT("average", "maximum", "minimum", "samples", "ts").ALIAS(
        "return", ("ts", 'timestamp')).FORMAT("return", ("ts", "YYYY-MM-DD HH24:MI:SS.MS")).WHERE(
        ["service", '=', service]).AND_WHERE(["monitor", '=', counter]).payload()
    result = await storage.query_tbl_with_payload('monitors', payload)
    response = {}
    if 'rows' in result:
        response = {"service": service, "monitors":{"monitor": counter}}
        response["monitors"]["values"] = result["rows"] if result["rows"] else []
    return web.json_response(response)

async def purge_all(request: web.Request) -> web.Response:
    """ DELETE all performance monitors

    :Example:
        curl -sX DELETE http://localhost:8081/foglamp/monitors
    """
    storage = connect.get_storage_async()
    result = await storage.delete_from_tbl("monitors", {})
    message = "Nothing to remove for service performance counters."
    if 'rows_affected' in result:
        if result['response'] == "deleted" and result['rows_affected']:
            message = "All Performance counters have been removed successfully."
    return web.json_response({"message": message})

async def purge_by_service(request: web.Request) -> web.Response:
    """ DELETE performance monitors for the given service

    :Example:
        curl -sX DELETE http://localhost:8081/foglamp/monitors/<SVC_NAME>
    """
    service = request.match_info.get('service', None)
    storage = connect.get_storage_async()
    payload = PayloadBuilder().WHERE(["service", '=', service]).payload()
    result = await storage.delete_from_tbl("monitors", payload)
    message = "Nothing to remove counters from '{}' service.".format(service)
    if 'rows_affected' in result:
        if result['response'] == "deleted" and result['rows_affected']:
            message = "Performance counters have been removed from '{}' service.".format(service)
    return web.json_response({"message": message})

async def purge_by_service_and_counter(request: web.Request) -> web.Response:
    """ DELETE performance monitors for the single counter for the single service

    :Example:
        curl -sX DELETE http://localhost:8081/foglamp/monitors/<SVC_NAME>/<COUNTER_NAME>
    """
    service = request.match_info.get('service', None)
    counter = request.match_info.get('counter', None)
    storage = connect.get_storage_async()
    payload = PayloadBuilder().WHERE(["service", '=', service]).AND_WHERE(
        ["monitor", '=', counter]).payload()
    result = await storage.delete_from_tbl("monitors", payload)
    message = "Nothing to remove '{}' counter from '{}' service.".format(counter, service)
    if 'rows_affected' in result:
        if result['response'] == "deleted" and result['rows_affected']:
            message = "Performance '{}' counter has been removed from '{}' service.".format(counter, service)
    return web.json_response({"message": message})
