# -*- coding: utf-8 -*-

# FOGLAMP_BEGIN
# See: http://foglamp.readthedocs.io/
# FOGLAMP_END

""" Test REST API endpoints with different user types """

import http.client
import json
import time
import pytest


__author__ = "Ashish Jabble"
__copyright__ = "Copyright (c) 2022 Dianomic Systems"
__license__ = "Apache 2.0"
__version__ = "${VERSION}"

TOKEN = None
VIEW_USERNAME = "view"
VIEW_PWD = "V!3w@1"
DATA_VIEW_USERNAME = "dataview"
DATA_VIEW_PWD = "DV!3w$"
CONTROL_USERNAME = "control"
CONTROL_PWD = "C0ntrol!"


@pytest.fixture
def change_to_auth_mandatory(foglamp_url, wait_time):
    # Wait for foglamp server to start
    time.sleep(wait_time)
    conn = http.client.HTTPConnection(foglamp_url)
    conn.request("PUT", '/foglamp/category/rest_api', json.dumps({"authentication": "mandatory"}))
    r = conn.getresponse()
    assert 200 == r.status
    r = r.read().decode()
    jdoc = json.loads(r)
    assert "mandatory" == jdoc['authentication']['value']

    conn.request("PUT", '/foglamp/restart', json.dumps({}))
    r = conn.getresponse()
    assert 200 == r.status
    r = r.read().decode()
    jdoc = json.loads(r)
    assert "FogLAMP restart has been scheduled." == jdoc['message']


def test_setup(reset_and_start_foglamp, change_to_auth_mandatory, foglamp_url, wait_time):
    time.sleep(wait_time * 3)
    conn = http.client.HTTPConnection(foglamp_url)
    # Admin login
    conn.request("POST", "/foglamp/login", json.dumps({"username": "admin", "password": "foglamp"}))
    r = conn.getresponse()
    assert 200 == r.status
    r = r.read().decode()
    jdoc = json.loads(r)
    assert "Logged in successfully." == jdoc['message']
    assert "token" in jdoc
    assert jdoc['admin']
    admin_token = jdoc["token"]
    # Create view user
    view_payload = {"username": VIEW_USERNAME, "password": VIEW_PWD, "role_id": 3, "real_name": "View",
                    "description": "Only to view the configuration"}
    conn.request("POST", "/foglamp/admin/user", body=json.dumps(view_payload), headers={"authorization": admin_token})
    r = conn.getresponse()
    assert 200 == r.status
    r = r.read().decode()
    jdoc = json.loads(r)
    assert "{} user has been created successfully.".format(VIEW_USERNAME) == jdoc["message"]
    # Create Data view user
    data_view_payload = {"username": DATA_VIEW_USERNAME, "password": DATA_VIEW_PWD, "role_id": 4,
                         "real_name": "DataView", "description": "Only read the data in buffer"}
    conn.request("POST", "/foglamp/admin/user", body=json.dumps(data_view_payload),
                 headers={"authorization": admin_token})
    r = conn.getresponse()
    assert 200 == r.status
    r = r.read().decode()
    jdoc = json.loads(r)
    assert "{} user has been created successfully.".format(DATA_VIEW_USERNAME) == jdoc["message"]
    # Create Control user
    control_payload = {"username": CONTROL_USERNAME, "password": CONTROL_PWD, "role_id": 5, "real_name": "Control",
                       "description": "Same as editor can do and also have access for control scripts and pipelines"}
    conn.request("POST", "/foglamp/admin/user", body=json.dumps(control_payload), headers={"authorization": admin_token})
    r = conn.getresponse()
    assert 200 == r.status
    r = r.read().decode()
    jdoc = json.loads(r)
    assert "{} user has been created successfully.".format(CONTROL_USERNAME) == jdoc["message"]


class TestAPIEndpointsWithViewUserType:
    def test_login(self, foglamp_url, wait_time):
        time.sleep(wait_time * 2)
        conn = http.client.HTTPConnection(foglamp_url)
        conn.request("POST", "/foglamp/login", json.dumps({"username": VIEW_USERNAME, "password": VIEW_PWD}))
        r = conn.getresponse()
        assert 200 == r.status
        r = r.read().decode()
        jdoc = json.loads(r)
        assert "Logged in successfully." == jdoc['message']
        assert "token" in jdoc
        assert not jdoc['admin']
        global TOKEN
        TOKEN = jdoc["token"]

    @pytest.mark.parametrize(("method", "route_path", "http_status_code"), [
        # common
        ("GET", "/foglamp/ping", 200), ("PUT", "/foglamp/shutdown", 403), ("PUT", "/foglamp/restart", 403),
        # health
        ("GET", "/foglamp/health/storage", 200), ("GET", "/foglamp/health/logging", 200),
        # user & roles
        ("GET", "/foglamp/user", 403), ("GET", "/foglamp/user?id=3", 200), ("GET", "/foglamp/user?id=2", 403),
        ("GET", "/foglamp/user?username={}".format(VIEW_USERNAME), 200),
        ("GET", "/foglamp/user?username={}".format(CONTROL_USERNAME), 403),
        ("GET", "/foglamp/user?id={}&username={}".format(3, VIEW_USERNAME), 200),
        ("GET", "/foglamp/user?username={}&id={}".format(VIEW_USERNAME, 3), 200),
        ("GET", "/foglamp/user?username=admin&id=1", 403),
        ("PUT", "/foglamp/user", 500), ("PUT", "/foglamp/user/1/password", 403), ("PUT", "/foglamp/user/3/password", 500),
        ("GET", "/foglamp/user/role", 403),
        # auth
        ("POST", "/foglamp/login", 403), ("PUT", "/foglamp/31/logout", 401),
        ("GET", "/foglamp/auth/ott", 200),
        # admin
        ("POST", "/foglamp/admin/user", 403), ("DELETE", "/foglamp/admin/3/delete", 403), ("PUT", "/foglamp/admin/3", 403),
        ("PUT", "/foglamp/admin/3/enable", 403), ("PUT", "/foglamp/admin/3/reset", 403),
        # category
        ("GET", "/foglamp/category", 200), ("POST", "/foglamp/category", 403), ("GET", "/foglamp/category/General", 200),
        ("PUT", "/foglamp/category/General", 403), ("DELETE", "/foglamp/category/General", 403),
        ("POST", "/foglamp/category/General/children", 403), ("GET", "/foglamp/category/General/children", 200),
        ("DELETE", "/foglamp/category/General/children/Advanced", 403),
        ("DELETE", "/foglamp/category/General/parent", 403),
        ("GET", "/foglamp/category/rest_api/allowPing", 200), ("PUT", "/foglamp/category/rest_api/allowPing", 403),
        ("DELETE", "/foglamp/category/rest_api/allowPing/value", 403),
        ("POST", "/foglamp/category/rest_api/allowPing/upload", 403),
        # schedule processes & schedules
        ("GET", "/foglamp/schedule/process", 200), ("POST", "/foglamp/schedule/process", 403),
        ("GET", "/foglamp/schedule/process/purge", 200),
        ("GET", "/foglamp/schedule", 200), ("POST", "/foglamp/schedule", 403), ("GET", "/foglamp/schedule/type", 200),
        ("GET", "/foglamp/schedule/2176eb68-7303-11e7-8cf7-a6006ad3dba0", 200),
        ("PUT", "/foglamp/schedule/2176eb68-7303-11e7-8cf7-a6006ad3dba0/enable", 403),
        ("PUT", "/foglamp/schedule/2176eb68-7303-11e7-8cf7-a6006ad3dba0/disable", 403),
        ("PUT", "/foglamp/schedule/enable", 403), ("PUT", "/foglamp/schedule/disable", 403),
        ("POST", "/foglamp/schedule/start/2176eb68-7303-11e7-8cf7-a6006ad3dba0", 403),
        ("PUT", "/foglamp/schedule/2176eb68-7303-11e7-8cf7-a6006ad3dba0", 403),
        ("DELETE", "/foglamp/schedule/2176eb68-7303-11e7-8cf7-a6006ad3dba0", 403),
        # tasks
        ("GET", "/foglamp/task", 200), ("GET", "/foglamp/task/state", 200), ("GET", "/foglamp/task/latest", 200),
        ("GET", "/foglamp/task/123", 404), ("PUT", "/foglamp/task/123/cancel", 403),
        ("POST", "/foglamp/scheduled/task", 403), ("DELETE", "/foglamp/scheduled/task/blah", 403),
        # service
        ("POST", "/foglamp/service", 403), ("GET", "/foglamp/service", 200), ("DELETE", "/foglamp/service/blah", 403),
        # ("GET", "/foglamp/service/available", 200), -- checked manually and commented out only to avoid apt-update
        ("GET", "/foglamp/service/installed", 200),
        ("PUT", "/foglamp/service/Southbound/blah/update", 403), ("POST", "/foglamp/service/blah/otp", 403),
        # south & north
        ("GET", "/foglamp/south", 200), ("GET", "/foglamp/north", 200),
        # asset browse
        ("GET", "/foglamp/asset", 200), ("GET", "/foglamp/asset/sinusoid", 200),
        ("GET", "/foglamp/asset/sinusoid/latest", 200),
        ("GET", "/foglamp/asset/sinusoid/summary", 404), ("GET", "/foglamp/asset/sinusoid/sinusoid", 200),
        ("GET", "/foglamp/asset/sinusoid/sinusoid/summary", 404), ("GET", "/foglamp/asset/sinusoid/sinusoid/series", 200),
        ("GET", "/foglamp/asset/sinusoid/bucket/1", 200), ("GET", "/foglamp/asset/sinusoid/sinusoid/bucket/1", 200),
        ("GET", "/foglamp/structure/asset", 200), ("DELETE", "/foglamp/asset", 403),
        ("DELETE", "/foglamp/asset/sinusoid", 403),
        # asset tracker
        ("GET", "/foglamp/track", 200), ("GET", "/foglamp/track/storage/assets", 200),
        ("PUT", "/foglamp/track/service/foo/asset/bar/event/Ingest", 403),
        # statistics
        ("GET", "/foglamp/statistics", 200), ("GET", "/foglamp/statistics/history", 200),
        ("GET", "/foglamp/statistics/rate?periods=1&statistics=FOO", 200),
        # audit trail
        ("POST", "/foglamp/audit", 403), ("GET", "/foglamp/audit", 200), ("GET", "/foglamp/audit/logcode", 200),
        ("GET", "/foglamp/audit/severity", 200),
        # backup & restore
        ("GET", "/foglamp/backup", 200), ("POST", "/foglamp/backup", 403), ("POST", "/foglamp/backup/upload", 403),
        ("GET", "/foglamp/backup/status", 200), ("GET", "/foglamp/backup/123", 404),
        ("DELETE", "/foglamp/backup/123", 403), ("GET", "/foglamp/backup/123/download", 403),
        ("PUT", "/foglamp/backup/123/restore", 403),
        # package update
        # ("GET", "/foglamp/update", 200), -- checked manually and commented out only to avoid apt-update run
        ("PUT", "/foglamp/update", 403),
        # certs store
        ("GET", "/foglamp/certificate", 200), ("POST", "/foglamp/certificate", 403),
        ("DELETE", "/foglamp/certificate/user", 403),
        # support bundle
        ("GET", "/foglamp/support", 200), ("GET", "/foglamp/support/foo", 403), ("POST", "/foglamp/support", 403),
        # syslogs & package logs
        ("GET", "/foglamp/syslog", 200), ("GET", "/foglamp/package/log", 200), ("GET", "/foglamp/package/log/foo", 400),
        ("GET", "/foglamp/package/install/status", 404),
        # plugins
        ("GET", "/foglamp/plugins/installed", 200),
        # ("GET", "/foglamp/plugins/available", 200), -- checked manually and commented out only to avoid apt-update
        ("POST", "/foglamp/plugins", 403), ("PUT", "/foglamp/plugins/south/sinusoid/update", 403),
        ("DELETE", "/foglamp/plugins/south/sinusoid", 403), ("GET", "/foglamp/service/foo/persist", 404),
        ("GET", "/foglamp/service/foo/plugin/omf/data", 404), ("POST", "/foglamp/service/foo/plugin/omf/data", 403),
        ("DELETE", "/foglamp/service/foo/plugin/omf/data", 403),
        # filters
        ("POST", "/foglamp/filter", 403), ("PUT", "/foglamp/filter/foo/pipeline", 403),
        ("GET", "/foglamp/filter/foo/pipeline", 404), ("GET", "/foglamp/filter/bar", 404), ("GET", "/foglamp/filter", 200),
        ("DELETE", "/foglamp/filter/foo/pipeline", 403), ("DELETE", "/foglamp/filter/bar", 403),
        # snapshots
        ("GET", "/foglamp/snapshot/plugins", 403), ("POST", "/foglamp/snapshot/plugins", 403),
        ("PUT", "/foglamp/snapshot/plugins/1", 403), ("DELETE", "/foglamp/snapshot/plugins/1", 403),
        ("GET", "/foglamp/snapshot/category", 403), ("POST", "/foglamp/snapshot/category", 403),
        ("PUT", "/foglamp/snapshot/category/1", 403), ("DELETE", "/foglamp/snapshot/category/1", 403),
        ("GET", "/foglamp/snapshot/schedule", 403), ("POST", "/foglamp/snapshot/schedule", 403),
        ("PUT", "/foglamp/snapshot/schedule/1", 403), ("DELETE", "/foglamp/snapshot/schedule/1", 403),
        # repository
        ("POST", "/foglamp/repository", 403),
        # ACL
        ("POST", "/foglamp/ACL", 403), ("GET", "/foglamp/ACL", 200), ("GET", "/foglamp/ACL/foo", 404),
        ("PUT", "/foglamp/ACL/foo", 403), ("DELETE", "/foglamp/ACL/foo", 403), ("PUT", "/foglamp/service/foo/ACL", 403),
        ("DELETE", "/foglamp/service/foo/ACL", 403),
        # control script
        ("POST", "/foglamp/control/script", 403), ("GET", "/foglamp/control/script", 200),
        ("GET", "/foglamp/control/script/foo", 404), ("PUT", "/foglamp/control/script/foo", 403),
        ("DELETE", "/foglamp/control/script/foo", 403), ("POST", "/foglamp/control/script/foo/schedule", 403),
        # control pipeline
        ("POST", "/foglamp/control/pipeline", 403), ("GET", "/foglamp/control/lookup", 200),
        ("GET", "/foglamp/control/pipeline", 200), ("GET", "/foglamp/control/pipeline/1", 404),
        ("PUT", "/foglamp/control/pipeline/1", 403), ("DELETE", "/foglamp/control/pipeline/1", 403),
        # python packages
        ("GET", "/foglamp/python/packages", 200), ("POST", "/foglamp/python/package", 403),
        # notification
        ("GET", "/foglamp/notification", 200), ("GET", "/foglamp/notification/plugin", 404),
        ("GET", "/foglamp/notification/type", 200), ("GET", "/foglamp/notification/N1", 400),
        ("POST", "/foglamp/notification", 403), ("PUT", "/foglamp/notification/N1", 403),
        ("DELETE", "/foglamp/notification/N1", 403), ("GET", "/foglamp/notification/N1/delivery", 404),
        ("POST", "/foglamp/notification/N1/delivery", 403), ("GET", "/foglamp/notification/N1/delivery/C1", 404),
        ("DELETE", "/foglamp/notification/N1/delivery/C1", 403),
        # performance monitors
        ("GET", "/foglamp/monitors", 200), ("GET", "/foglamp/monitors/SVC", 200),
        ("GET", "/foglamp/monitors/Svc/Counter", 200), ("DELETE", "/foglamp/monitors", 403),
        ("DELETE", "/foglamp/monitors/SVC", 403), ("DELETE", "/foglamp/monitors/Svc/Counter", 403),
        # alerts
        ("GET", "/foglamp/alert", 200), ("DELETE", "/foglamp/alert", 403), ("DELETE", "/foglamp/alert/blah", 403)
    ])
    def test_endpoints(self, foglamp_url, method, route_path, http_status_code, storage_plugin):
        conn = http.client.HTTPConnection(foglamp_url)
        conn.request(method, route_path, headers={"authorization": TOKEN})
        r = conn.getresponse()
        assert http_status_code == r.status
        r.read().decode()

    def test_logout_me(self, foglamp_url):
        conn = http.client.HTTPConnection(foglamp_url)
        conn.request("PUT", '/foglamp/logout', headers={"authorization": TOKEN})
        r = conn.getresponse()
        assert 200 == r.status
        r = r.read().decode()
        jdoc = json.loads(r)
        assert jdoc['logout']


class TestAPIEndpointsWithDataViewUserType:
    def test_login(self, foglamp_url, wait_time):
        time.sleep(wait_time * 2)
        conn = http.client.HTTPConnection(foglamp_url)
        conn.request("POST", "/foglamp/login", json.dumps({"username": DATA_VIEW_USERNAME, "password": DATA_VIEW_PWD}))
        r = conn.getresponse()
        assert 200 == r.status
        r = r.read().decode()
        jdoc = json.loads(r)
        assert "Logged in successfully." == jdoc['message']
        assert "token" in jdoc
        assert not jdoc['admin']
        global TOKEN
        TOKEN = jdoc["token"]

    @pytest.mark.parametrize(("method", "route_path", "http_status_code"), [
        # common
        ("GET", "/foglamp/ping", 200), ("PUT", "/foglamp/shutdown", 403), ("PUT", "/foglamp/restart", 403),
        # health
        ("GET", "/foglamp/health/storage", 403), ("GET", "/foglamp/health/logging", 403),
        # user & roles
        ("GET", "/foglamp/user", 403), ("GET", "/foglamp/user?id=4", 200), ("GET", "/foglamp/user?id=1", 403),
        ("GET", "/foglamp/user?username={}".format(DATA_VIEW_USERNAME), 200), ("GET", "/foglamp/user?username=user", 403),
        ("GET", "/foglamp/user?id={}&username={}".format(4, DATA_VIEW_USERNAME), 200),
        ("GET", "/foglamp/user?id=1&username=admin", 403),
        ("GET", "/foglamp/user?username={}&id={}".format(DATA_VIEW_USERNAME, 4), 200),
        ("PUT", "/foglamp/user", 500), ("PUT", "/foglamp/user/1/password", 403), ("PUT", "/foglamp/user/4/password", 500),
        ("GET", "/foglamp/user/role", 403),
        # auth
        ("POST", "/foglamp/login", 403), ("PUT", "/foglamp/31/logout", 401),
        ("GET", "/foglamp/auth/ott", 403),
        # admin
        ("POST", "/foglamp/admin/user", 403), ("DELETE", "/foglamp/admin/3/delete", 403), ("PUT", "/foglamp/admin/3", 403),
        ("PUT", "/foglamp/admin/3/enable", 403), ("PUT", "/foglamp/admin/3/reset", 403),
        # category
        ("GET", "/foglamp/category", 403), ("POST", "/foglamp/category", 403), ("GET", "/foglamp/category/General", 403),
        ("PUT", "/foglamp/category/General", 403), ("DELETE", "/foglamp/category/General", 403),
        ("POST", "/foglamp/category/General/children", 403), ("GET", "/foglamp/category/General/children", 403),
        ("DELETE", "/foglamp/category/General/children/Advanced", 403),
        ("DELETE", "/foglamp/category/General/parent", 403),
        ("GET", "/foglamp/category/rest_api/allowPing", 403), ("PUT", "/foglamp/category/rest_api/allowPing", 403),
        ("DELETE", "/foglamp/category/rest_api/allowPing/value", 403),
        ("POST", "/foglamp/category/rest_api/allowPing/upload", 403),
        # schedule processes & schedules
        ("GET", "/foglamp/schedule/process", 403), ("POST", "/foglamp/schedule/process", 403),
        ("GET", "/foglamp/schedule/process/purge", 403),
        ("GET", "/foglamp/schedule", 403), ("POST", "/foglamp/schedule", 403), ("GET", "/foglamp/schedule/type", 403),
        ("GET", "/foglamp/schedule/2176eb68-7303-11e7-8cf7-a6006ad3dba0", 403),
        ("PUT", "/foglamp/schedule/2176eb68-7303-11e7-8cf7-a6006ad3dba0/enable", 403),
        ("PUT", "/foglamp/schedule/2176eb68-7303-11e7-8cf7-a6006ad3dba0/disable", 403),
        ("PUT", "/foglamp/schedule/enable", 403), ("PUT", "/foglamp/schedule/disable", 403),
        ("POST", "/foglamp/schedule/start/2176eb68-7303-11e7-8cf7-a6006ad3dba0", 403),
        ("PUT", "/foglamp/schedule/2176eb68-7303-11e7-8cf7-a6006ad3dba0", 403),
        ("DELETE", "/foglamp/schedule/2176eb68-7303-11e7-8cf7-a6006ad3dba0", 403),
        # tasks
        ("GET", "/foglamp/task", 403), ("GET", "/foglamp/task/state", 403), ("GET", "/foglamp/task/latest", 403),
        ("GET", "/foglamp/task/123", 403), ("PUT", "/foglamp/task/123/cancel", 403),
        ("POST", "/foglamp/scheduled/task", 403), ("DELETE", "/foglamp/scheduled/task/blah", 403),
        # service
        ("POST", "/foglamp/service", 403), ("GET", "/foglamp/service", 200), ("DELETE", "/foglamp/service/blah", 403),
        ("GET", "/foglamp/service/available", 403), ("GET", "/foglamp/service/installed", 403),
        ("PUT", "/foglamp/service/Southbound/blah/update", 403), ("POST", "/foglamp/service/blah/otp", 403),
        # south & north
        ("GET", "/foglamp/south", 403), ("GET", "/foglamp/north", 403),
        # asset browse
        ("GET", "/foglamp/asset", 200), ("GET", "/foglamp/asset/sinusoid", 200),
        ("GET", "/foglamp/asset/sinusoid/latest", 200),
        ("GET", "/foglamp/asset/sinusoid/summary", 404), ("GET", "/foglamp/asset/sinusoid/sinusoid", 200),
        ("GET", "/foglamp/asset/sinusoid/sinusoid/summary", 404), ("GET", "/foglamp/asset/sinusoid/sinusoid/series", 200),
        ("GET", "/foglamp/asset/sinusoid/bucket/1", 200), ("GET", "/foglamp/asset/sinusoid/sinusoid/bucket/1", 200),
        ("GET", "/foglamp/structure/asset", 403), ("DELETE", "/foglamp/asset", 403),
        ("DELETE", "/foglamp/asset/sinusoid", 403),
        # asset tracker
        ("GET", "/foglamp/track", 403), ("GET", "/foglamp/track/storage/assets", 403),
        ("PUT", "/foglamp/track/service/foo/asset/bar/event/Ingest", 403),
        # statistics
        ("GET", "/foglamp/statistics", 200), ("GET", "/foglamp/statistics/history", 200),
        ("GET", "/foglamp/statistics/rate?periods=1&statistics=FOO", 200),
        # audit trail
        ("POST", "/foglamp/audit", 403), ("GET", "/foglamp/audit", 403), ("GET", "/foglamp/audit/logcode", 403),
        ("GET", "/foglamp/audit/severity", 403),
        # backup & restore
        ("GET", "/foglamp/backup", 403), ("POST", "/foglamp/backup", 403), ("POST", "/foglamp/backup/upload", 403),
        ("GET", "/foglamp/backup/status", 403), ("GET", "/foglamp/backup/123", 403),
        ("DELETE", "/foglamp/backup/123", 403), ("GET", "/foglamp/backup/123/download", 403),
        ("PUT", "/foglamp/backup/123/restore", 403),
        # package update
        # ("GET", "/foglamp/update", 200), -- checked manually and commented out only to avoid apt-update
        ("PUT", "/foglamp/update", 403),
        # certs store
        ("GET", "/foglamp/certificate", 403), ("POST", "/foglamp/certificate", 403),
        ("DELETE", "/foglamp/certificate/user", 403),
        # support bundle
        ("GET", "/foglamp/support", 403), ("GET", "/foglamp/support/foo", 403), ("POST", "/foglamp/support", 403),
        # syslogs & package logs
        ("GET", "/foglamp/syslog", 403), ("GET", "/foglamp/package/log", 403), ("GET", "/foglamp/package/log/foo", 403),
        ("GET", "/foglamp/package/install/status", 403),
        # plugins
        ("GET", "/foglamp/plugins/installed", 403), ("GET", "/foglamp/plugins/available", 403),
        ("POST", "/foglamp/plugins", 403), ("PUT", "/foglamp/plugins/south/sinusoid/update", 403),
        ("DELETE", "/foglamp/plugins/south/sinusoid", 403), ("GET", "/foglamp/service/foo/persist", 403),
        ("GET", "/foglamp/service/foo/plugin/omf/data", 403), ("POST", "/foglamp/service/foo/plugin/omf/data", 403),
        ("DELETE", "/foglamp/service/foo/plugin/omf/data", 403),
        # filters
        ("POST", "/foglamp/filter", 403), ("PUT", "/foglamp/filter/foo/pipeline", 403),
        ("GET", "/foglamp/filter/foo/pipeline", 403), ("GET", "/foglamp/filter/bar", 403), ("GET", "/foglamp/filter", 403),
        ("DELETE", "/foglamp/filter/foo/pipeline", 403), ("DELETE", "/foglamp/filter/bar", 403),
        # snapshots
        ("GET", "/foglamp/snapshot/plugins", 403), ("POST", "/foglamp/snapshot/plugins", 403),
        ("PUT", "/foglamp/snapshot/plugins/1", 403), ("DELETE", "/foglamp/snapshot/plugins/1", 403),
        ("GET", "/foglamp/snapshot/category", 403), ("POST", "/foglamp/snapshot/category", 403),
        ("PUT", "/foglamp/snapshot/category/1", 403), ("DELETE", "/foglamp/snapshot/category/1", 403),
        ("GET", "/foglamp/snapshot/schedule", 403), ("POST", "/foglamp/snapshot/schedule", 403),
        ("PUT", "/foglamp/snapshot/schedule/1", 403), ("DELETE", "/foglamp/snapshot/schedule/1", 403),
        # repository
        ("POST", "/foglamp/repository", 403),
        # ACL
        ("POST", "/foglamp/ACL", 403), ("GET", "/foglamp/ACL", 403), ("GET", "/foglamp/ACL/foo", 403),
        ("PUT", "/foglamp/ACL/foo", 403), ("DELETE", "/foglamp/ACL/foo", 403), ("PUT", "/foglamp/service/foo/ACL", 403),
        ("DELETE", "/foglamp/service/foo/ACL", 403),
        # control script
        ("POST", "/foglamp/control/script", 403), ("GET", "/foglamp/control/script", 403),
        ("GET", "/foglamp/control/script/foo", 403), ("PUT", "/foglamp/control/script/foo", 403),
        ("DELETE", "/foglamp/control/script/foo", 403), ("POST", "/foglamp/control/script/foo/schedule", 403),
        # control pipeline
        ("POST", "/foglamp/control/pipeline", 403), ("GET", "/foglamp/control/lookup", 403),
        ("GET", "/foglamp/control/pipeline", 403), ("GET", "/foglamp/control/pipeline/1", 403),
        ("PUT", "/foglamp/control/pipeline/1", 403), ("DELETE", "/foglamp/control/pipeline/1", 403),
        # python packages
        ("GET", "/foglamp/python/packages", 403), ("POST", "/foglamp/python/package", 403),
        # notification
        ("GET", "/foglamp/notification", 403), ("GET", "/foglamp/notification/plugin", 403),
        ("GET", "/foglamp/notification/type", 403), ("GET", "/foglamp/notification/N1", 403),
        ("POST", "/foglamp/notification", 403), ("PUT", "/foglamp/notification/N1", 403),
        ("DELETE", "/foglamp/notification/N1", 403), ("GET", "/foglamp/notification/N1/delivery", 403),
        ("POST", "/foglamp/notification/N1/delivery", 403), ("GET", "/foglamp/notification/N1/delivery/C1", 403),
        ("DELETE", "/foglamp/notification/N1/delivery/C1", 403),
        # performance monitors
        ("GET", "/foglamp/monitors", 403), ("GET", "/foglamp/monitors/SVC", 403),
        ("GET", "/foglamp/monitors/Svc/Counter", 403), ("DELETE", "/foglamp/monitors", 403),
        ("DELETE", "/foglamp/monitors/SVC", 403), ("DELETE", "/foglamp/monitors/Svc/Counter", 403),
        # alerts
        ("GET", "/foglamp/alert", 403), ("DELETE", "/foglamp/alert", 403), ("DELETE", "/foglamp/alert/blah", 403)
    ])
    def test_endpoints(self, foglamp_url, method, route_path, http_status_code, storage_plugin):
        conn = http.client.HTTPConnection(foglamp_url)
        conn.request(method, route_path, headers={"authorization": TOKEN})
        r = conn.getresponse()
        assert http_status_code == r.status
        r.read().decode()

    def test_logout_me(self, foglamp_url):
        conn = http.client.HTTPConnection(foglamp_url)
        conn.request("PUT", '/foglamp/logout', headers={"authorization": TOKEN})
        r = conn.getresponse()
        assert 200 == r.status
        r = r.read().decode()
        jdoc = json.loads(r)
        assert jdoc['logout']


class TestAPIEndpointsWithControlUserType:
    def test_login(self, foglamp_url, wait_time):
        time.sleep(wait_time * 2)
        conn = http.client.HTTPConnection(foglamp_url)
        conn.request("POST", "/foglamp/login", json.dumps({"username": CONTROL_USERNAME,
                                                          "password": CONTROL_PWD}))
        r = conn.getresponse()
        assert 200 == r.status
        r = r.read().decode()
        jdoc = json.loads(r)
        assert "Logged in successfully." == jdoc['message']
        assert "token" in jdoc
        assert not jdoc['admin']
        global TOKEN
        TOKEN = jdoc["token"]

    @pytest.mark.parametrize(("method", "route_path", "http_status_code"), [
        # common
        ("GET", "/foglamp/ping", 200),  # ("PUT", "/foglamp/shutdown", 200), ("PUT", "/foglamp/restart", 200),
        # health
        ("GET", "/foglamp/health/storage", 200), ("GET", "/foglamp/health/logging", 200),
        # user & roles
        ("GET", "/foglamp/user", 200), ("GET", "/foglamp/user?id=5", 200), ("GET", "/foglamp/user?id=1", 200),
        ("GET", "/foglamp/user?username={}".format(CONTROL_USERNAME), 200), ("GET", "/foglamp/user?username=admin", 200),
        ("GET", "/foglamp/user?id={}&username={}".format(5, CONTROL_USERNAME), 200),
        ("GET", "/foglamp/user?username={}&id={}".format(CONTROL_USERNAME, 5), 200),
        ("GET", "/foglamp/user?username={}&id={}".format(VIEW_USERNAME, 3), 200),
        ("GET", "/foglamp/user?id={}&username={}".format(4, DATA_VIEW_USERNAME), 200),
        ("PUT", "/foglamp/user", 500), ("PUT", "/foglamp/user/1/password", 401), ("PUT", "/foglamp/user/5/password", 500),
        ("GET", "/foglamp/user/role", 403),
        # auth
        ("POST", "/foglamp/login", 500), ("PUT", "/foglamp/31/logout", 401),
        ("GET", "/foglamp/auth/ott", 200),
        # admin
        ("POST", "/foglamp/admin/user", 403), ("DELETE", "/foglamp/admin/3/delete", 403), ("PUT", "/foglamp/admin/3", 403),
        ("PUT", "/foglamp/admin/3/enable", 403), ("PUT", "/foglamp/admin/3/reset", 403),
        # category
        ("GET", "/foglamp/category", 200), ("POST", "/foglamp/category", 400), ("GET", "/foglamp/category/General", 200),
        ("PUT", "/foglamp/category/General", 400), ("DELETE", "/foglamp/category/General", 400),
        ("POST", "/foglamp/category/General/children", 500), ("GET", "/foglamp/category/General/children", 200),
        ("DELETE", "/foglamp/category/General/children/Advanced", 200),
        ("DELETE", "/foglamp/category/General/parent", 200),
        ("GET", "/foglamp/category/rest_api/allowPing", 200), ("PUT", "/foglamp/category/rest_api/allowPing", 500),
        ("DELETE", "/foglamp/category/rest_api/allowPing/value", 200),
        ("POST", "/foglamp/category/rest_api/allowPing/upload", 400),
        # schedule processes & schedules
        ("GET", "/foglamp/schedule/process", 200), ("POST", "/foglamp/schedule/process", 500),
        ("GET", "/foglamp/schedule/process/purge", 200),
        ("GET", "/foglamp/schedule", 200), ("POST", "/foglamp/schedule", 400), ("GET", "/foglamp/schedule/type", 200),
        ("GET", "/foglamp/schedule/2176eb68-7303-11e7-8cf7-a6006ad3dba0", 200),
        ("PUT", "/foglamp/schedule/2176eb68-7303-11e7-8cf7-a6006ad3dba0/enable", 200),
        ("PUT", "/foglamp/schedule/2176eb68-7303-11e7-8cf7-a6006ad3dba0/disable", 200),
        ("PUT", "/foglamp/schedule/enable", 404), ("PUT", "/foglamp/schedule/disable", 404),
        ("POST", "/foglamp/schedule/start/2176eb68-7303-11e7-8cf7-a6006ad3dba0", 200),
        ("PUT", "/foglamp/schedule/2176eb68-7303-11e7-8cf7-a6006ad3dba0", 400),
        ("DELETE", "/foglamp/schedule/d1631422-9ec6-11e7-abc4-cec278b6b50a", 200),
        # tasks
        ("GET", "/foglamp/task", 200), ("GET", "/foglamp/task/state", 200), ("GET", "/foglamp/task/latest", 200),
        ("GET", "/foglamp/task/123", 404), ("PUT", "/foglamp/task/123/cancel", 404),
        ("POST", "/foglamp/scheduled/task", 400), ("DELETE", "/foglamp/scheduled/task/blah", 404),
        # service
        ("POST", "/foglamp/service", 400), ("GET", "/foglamp/service", 200), ("DELETE", "/foglamp/service/blah", 404),
        # ("GET", "/foglamp/service/available", 200), -- checked manually and commented out only to avoid apt-update
        ("GET", "/foglamp/service/installed", 200),
        ("PUT", "/foglamp/service/Southbound/blah/update", 400), ("POST", "/foglamp/service/blah/otp", 403),
        # south & north
        ("GET", "/foglamp/south", 200), ("GET", "/foglamp/north", 200),
        # asset browse
        ("GET", "/foglamp/asset", 200), ("GET", "/foglamp/asset/sinusoid", 200),
        ("GET", "/foglamp/asset/sinusoid/latest", 200),
        ("GET", "/foglamp/asset/sinusoid/summary", 404), ("GET", "/foglamp/asset/sinusoid/sinusoid", 200),
        ("GET", "/foglamp/asset/sinusoid/sinusoid/summary", 404), ("GET", "/foglamp/asset/sinusoid/sinusoid/series", 200),
        ("GET", "/foglamp/asset/sinusoid/bucket/1", 200), ("GET", "/foglamp/asset/sinusoid/sinusoid/bucket/1", 200),
        ("GET", "/foglamp/structure/asset", 200), ("DELETE", "/foglamp/asset", 200),
        ("DELETE", "/foglamp/asset/sinusoid", 200),
        # asset tracker
        ("GET", "/foglamp/track", 200), ("GET", "/foglamp/track/storage/assets", 200),
        ("PUT", "/foglamp/track/service/foo/asset/bar/event/Ingest", 404),
        # statistics
        ("GET", "/foglamp/statistics", 200), ("GET", "/foglamp/statistics/history", 200),
        ("GET", "/foglamp/statistics/rate?periods=1&statistics=FOO", 200),
        # audit trail
        ("POST", "/foglamp/audit", 500), ("GET", "/foglamp/audit", 200), ("GET", "/foglamp/audit/logcode", 200),
        ("GET", "/foglamp/audit/severity", 200),
        # backup & restore
        ("GET", "/foglamp/backup", 200), ("POST", "/foglamp/backup", 403),
        ("POST", "/foglamp/backup/upload", 403),
        ("GET", "/foglamp/backup/status", 200), ("GET", "/foglamp/backup/123", 404),
        ("DELETE", "/foglamp/backup/123", 403), ("GET", "/foglamp/backup/123/download", 403),
        ("PUT", "/foglamp/backup/123/restore", 403),
        # package update
        # ("GET", "/foglamp/update", 200), -- checked manually and commented out only to avoid apt-update run
        # ("PUT", "/foglamp/update", 200), -- checked manually
        # certs store
        ("GET", "/foglamp/certificate", 200), ("POST", "/foglamp/certificate", 400),
        ("DELETE", "/foglamp/certificate/user", 403),
        # support bundle
        ("GET", "/foglamp/support", 200), ("GET", "/foglamp/support/foo", 403), ("POST", "/foglamp/support", 403),
        # syslogs & package logs
        ("GET", "/foglamp/syslog", 200), ("GET", "/foglamp/package/log", 200), ("GET", "/foglamp/package/log/foo", 400),
        ("GET", "/foglamp/package/install/status", 404),
        # plugins
        ("GET", "/foglamp/plugins/installed", 200),
        # ("GET", "/foglamp/plugins/available", 200), -- checked manually and commented out only to avoid apt operations
        # ("PUT", "/foglamp/plugins/south/sinusoid/update", 200),
        # ("DELETE", "/foglamp/plugins/south/sinusoid", 404),
        ("POST", "/foglamp/plugins", 400), ("GET", "/foglamp/service/foo/persist", 404),
        ("GET", "/foglamp/service/foo/plugin/omf/data", 404), ("POST", "/foglamp/service/foo/plugin/omf/data", 404),
        ("DELETE", "/foglamp/service/foo/plugin/omf/data", 404),
        # filters
        ("POST", "/foglamp/filter", 404), ("PUT", "/foglamp/filter/foo/pipeline", 404),
        ("GET", "/foglamp/filter/foo/pipeline", 404), ("GET", "/foglamp/filter/bar", 404), ("GET", "/foglamp/filter", 200),
        ("DELETE", "/foglamp/filter/foo/pipeline", 500), ("DELETE", "/foglamp/filter/bar", 404),
        # snapshots
        ("GET", "/foglamp/snapshot/plugins", 403), ("POST", "/foglamp/snapshot/plugins", 403),
        ("PUT", "/foglamp/snapshot/plugins/1", 403), ("DELETE", "/foglamp/snapshot/plugins/1", 403),
        ("GET", "/foglamp/snapshot/category", 403), ("POST", "/foglamp/snapshot/category", 403),
        ("PUT", "/foglamp/snapshot/category/1", 403), ("DELETE", "/foglamp/snapshot/category/1", 403),
        ("GET", "/foglamp/snapshot/schedule", 403), ("POST", "/foglamp/snapshot/schedule", 403),
        ("PUT", "/foglamp/snapshot/schedule/1", 403), ("DELETE", "/foglamp/snapshot/schedule/1", 403),
        # repository
        ("POST", "/foglamp/repository", 400),
        # ACL
        ("POST", "/foglamp/ACL", 403), ("GET", "/foglamp/ACL", 200), ("GET", "/foglamp/ACL/foo", 404),
        ("PUT", "/foglamp/ACL/foo", 403), ("DELETE", "/foglamp/ACL/foo", 403), ("PUT", "/foglamp/service/foo/ACL", 403),
        ("DELETE", "/foglamp/service/foo/ACL", 403),
        # control script
        ("POST", "/foglamp/control/script", 400), ("GET", "/foglamp/control/script", 200),
        ("GET", "/foglamp/control/script/foo", 404), ("PUT", "/foglamp/control/script/foo", 400),
        ("DELETE", "/foglamp/control/script/foo", 404), ("POST", "/foglamp/control/script/foo/schedule", 404),
        # control pipeline
        ("POST", "/foglamp/control/pipeline", 400), ("GET", "/foglamp/control/lookup", 200),
        ("GET", "/foglamp/control/pipeline", 200), ("GET", "/foglamp/control/pipeline/1", 404),
        ("PUT", "/foglamp/control/pipeline/1", 404), ("DELETE", "/foglamp/control/pipeline/1", 404),
        # python packages
        ("GET", "/foglamp/python/packages", 200), ("POST", "/foglamp/python/package", 500),
        # notification
        ("GET", "/foglamp/notification", 200), ("GET", "/foglamp/notification/plugin", 404),
        ("GET", "/foglamp/notification/type", 200), ("GET", "/foglamp/notification/N1", 400),
        ("POST", "/foglamp/notification", 404), ("PUT", "/foglamp/notification/N1", 404),
        ("DELETE", "/foglamp/notification/N1", 404), ("GET", "/foglamp/notification/N1/delivery", 404),
        ("POST", "/foglamp/notification/N1/delivery", 400), ("GET", "/foglamp/notification/N1/delivery/C1", 404),
        ("DELETE", "/foglamp/notification/N1/delivery/C1", 404),
        # performance monitors
        ("GET", "/foglamp/monitors", 200), ("GET", "/foglamp/monitors/SVC", 200),
        ("GET", "/foglamp/monitors/Svc/Counter", 200), ("DELETE", "/foglamp/monitors", 200),
        ("DELETE", "/foglamp/monitors/SVC", 200), ("DELETE", "/foglamp/monitors/Svc/Counter", 200),
        # alerts
        ("GET", "/foglamp/alert", 200), ("DELETE", "/foglamp/alert", 200), ("DELETE", "/foglamp/alert/blah", 404)
    ])
    def test_endpoints(self, foglamp_url, method, route_path, http_status_code, storage_plugin):
        conn = http.client.HTTPConnection(foglamp_url)
        conn.request(method, route_path, headers={"authorization": TOKEN})
        r = conn.getresponse()
        assert http_status_code == r.status
        r.read().decode()

    def test_logout_me(self, foglamp_url):
        conn = http.client.HTTPConnection(foglamp_url)
        conn.request("PUT", '/foglamp/logout', headers={"authorization": TOKEN})
        r = conn.getresponse()
        assert 200 == r.status
        r = r.read().decode()
        jdoc = json.loads(r)
        assert jdoc['logout']
