# -*- coding: utf-8 -*-

# FOGLAMP_BEGIN
# See: http://foglamp.readthedocs.io/
# FOGLAMP_END

""" Test Statistics & Statistics history REST API """

import os
import subprocess
import http.client
import json
import time
from collections import Counter
import pytest
import utils


__author__ = "Ashish Jabble"
__copyright__ = "Copyright (c) 2019 Dianomic Systems"
__license__ = "Apache 2.0"
__version__ = "${VERSION}"


STATS_KEYS = {'DISCARDED', 'PURGED', 'READINGS', 'UNSENT', 'UNSNPURGED', 'BUFFERED'}
STATS_HISTORY_KEYS = {'history_ts'}
STATS_HISTORY_KEYS.update(STATS_KEYS)
TEMPLATE_NAME = "template.json"
SENSOR_VALUE = 10
PLUGIN_NAME = "coap"
ASSET_NAME = "COAP"


@pytest.fixture
def start_south_coap(add_south, remove_data_file, remove_directories, south_branch,
                     foglamp_url, south_plugin=PLUGIN_NAME, asset_name=ASSET_NAME):
    # Define the template file for fogbench
    fogbench_template_path = os.path.join(
        os.path.expandvars('${FOGLAMP_ROOT}'), 'data/{}'.format(TEMPLATE_NAME))
    with open(fogbench_template_path, "w") as f:
        f.write(
            '[{"name": "%s", "sensor_values": '
            '[{"name": "sensor", "type": "number", "min": %d, "max": %d, "precision": 0}]}]' % (
                asset_name, SENSOR_VALUE, SENSOR_VALUE))

    add_south(south_plugin, south_branch, foglamp_url, service_name=PLUGIN_NAME)

    yield start_south_coap

    # Cleanup code that runs after the caller test is over
    remove_data_file(fogbench_template_path)
    remove_directories("/tmp/foglamp-south-{}".format(south_plugin))


class TestStatistics:

    def test_cleanup(self, reset_and_start_foglamp):
        # TODO: Remove this workaround
        # Use better setup & teardown methods
        pass

    def test_default_statistics(self, foglamp_url):
        conn = http.client.HTTPConnection(foglamp_url)
        conn.request("GET", '/foglamp/statistics')
        r = conn.getresponse()
        assert 200 == r.status
        r = r.read().decode()
        jdoc = json.loads(r)
        assert len(jdoc), "No data found"
        assert len(STATS_KEYS) == len(jdoc)
        keys = [key['key'] for key in jdoc]
        assert Counter(STATS_KEYS) == Counter(keys)

    def test_default_statistics_history(self, foglamp_url):
        conn = http.client.HTTPConnection(foglamp_url)
        conn.request("GET", '/foglamp/statistics/history')
        r = conn.getresponse()
        assert 200 == r.status
        r = r.read().decode()
        jdoc = json.loads(r)
        assert len(jdoc), "No data found"
        assert 15 == jdoc['interval']
        assert {} == jdoc['statistics'][0]

    def test_statistics_history_with_stats_collector_schedule(self, foglamp_url, wait_time, retries):
        conn = http.client.HTTPConnection(foglamp_url)
        for i in range (0,retries):
            # wait for sometime for stats collector schedule to start
            time.sleep(wait_time)

            conn.request("GET", '/foglamp/statistics/history')
            r = conn.getresponse()
            assert 200 == r.status
            r = r.read().decode()
            jdoc = json.loads(r)
            assert len(jdoc), "No data found"
            assert 15 == jdoc['interval']
            assert 1 == len(jdoc['statistics'])
            statistics = jdoc['statistics'][0]
            if i <= retries and Counter(STATS_HISTORY_KEYS) != Counter(statistics.keys()): continue
            assert Counter(STATS_HISTORY_KEYS) == Counter(statistics.keys())
            assert utils.validate_date_format(statistics['history_ts']) is True, "History Timestamp format mismatched."

            del statistics['history_ts']
            assert Counter([0, 0, 0, 0, 0, 0]) == Counter(statistics.values())
            break

    @pytest.mark.parametrize("request_params, keys", [
        ('', STATS_HISTORY_KEYS),
        ('?limit=1', STATS_HISTORY_KEYS),
        ('?key=READINGS', {'history_ts', 'READINGS'}),
        ('?key=READINGS&limit=1', {'history_ts', 'READINGS'}),
        ('?key=READINGS&limit=0', {}),
    ])
    def test_statistics_history_with_params(self, foglamp_url, request_params, keys):
        conn = http.client.HTTPConnection(foglamp_url)
        conn.request("GET", '/foglamp/statistics/history{}'.format(request_params))
        r = conn.getresponse()
        assert 200 == r.status
        r = r.read().decode()
        jdoc = json.loads(r)
        assert len(jdoc), "No data found"
        assert 15 == jdoc['interval']
        assert 1 == len(jdoc['statistics'])
        assert Counter(keys) == Counter(jdoc['statistics'][0].keys())
        if keys:
            assert utils.validate_date_format(jdoc['statistics'][0]['history_ts']) is True, \
                "History Timestamp format mismatched."

    def test_statistics_history_with_service_enabled(self, start_south_coap, foglamp_url, wait_time):
        # Allow CoAP listener to start
        time.sleep(wait_time)

        # ingest one reading via fogbench
        subprocess.run(["cd $FOGLAMP_ROOT/extras/python; python3 -m fogbench -t ../../data/{}; cd -"
                       .format(TEMPLATE_NAME)], shell=True, check=True)
        # Let the readings to be Ingressed
        time.sleep(wait_time)

        # verify stats
        conn = http.client.HTTPConnection(foglamp_url)
        conn.request("GET", '/foglamp/statistics')
        r = conn.getresponse()
        assert 200 == r.status
        r = r.read().decode()
        jdoc = json.loads(r)
        assert len(jdoc), "No data found"
        stats = utils.serialize_stats_map(jdoc)
        assert 1 == stats[ASSET_NAME.upper()]
        assert 1 == stats['READINGS']

        # Allow stats collector schedule to run i.e. by default 15s
        time.sleep(wait_time * 3)

        # check stats history
        conn.request("GET", '/foglamp/statistics/history')
        r = conn.getresponse()
        assert 200 == r.status
        r = r.read().decode()
        jdoc = json.loads(r)
        assert len(jdoc), "No data found"
        stats_history = jdoc['statistics']

        # READINGS & ASSET_NAME keys and verify no duplicate entry found with value 1
        read = [r['READINGS'] for r in stats_history]
        assert 1 in read
        assert 1 == read.count(1)
        asset_stats_history = [a for a in stats_history if ASSET_NAME.upper() in a.keys()]
        assert any(ash[ASSET_NAME.upper()] == 1 for ash in asset_stats_history), "Failed to find statistics history " \
                                                                                 "record for " + ASSET_NAME.upper()

        # verify stats history by READINGS key only
        conn.request("GET", '/foglamp/statistics/history?key={}'.format('READINGS'))
        r = conn.getresponse()
        assert 200 == r.status
        r = r.read().decode()
        jdoc = json.loads(r)
        assert len(jdoc), "No data found"
        for stat in jdoc['statistics']:
            assert utils.validate_date_format(stat['history_ts']) is True, "History Timestamp format mismatched."

        read = [r['READINGS'] for r in jdoc['statistics']]
        assert 1 in read
        assert 1 == read.count(1)

        # verify stats history by ASSET_NAME key only
        conn.request("GET", '/foglamp/statistics/history?key={}'.format(ASSET_NAME.upper()))
        r = conn.getresponse()
        assert 200 == r.status
        r = r.read().decode()
        jdoc = json.loads(r)
        assert len(jdoc), "No data found"
        asset = [a[ASSET_NAME.upper()] for a in jdoc['statistics']]
        assert 1 in asset
        assert 1 == asset.count(1)

