# FOGLAMP_BEGIN
# See: http://foglamp.readthedocs.io/
# FOGLAMP_END

""" Test data availability notification rule system tests:
        Creates notification instance with data availability rule
        and notify asset plugin for triggering the notifications based on CONCH.
"""

__author__ = "Yash Tatkondawar"
__copyright__ = "Copyright (c) 2023 Dianomic Systems, Inc."


import os
import subprocess
import time
import urllib.parse
import json
from pathlib import Path
import http
from datetime import datetime
import pytest
import utils
from pytest import PKG_MGR

# This  gives the path of directory where foglamp is cloned. test_file < packages < python < system < tests < ROOT
PROJECT_ROOT = Path(__file__).parent.parent.parent.parent.parent
SCRIPTS_DIR_ROOT = "{}/tests/system/python/packages/data/".format(PROJECT_ROOT)
FOGLAMP_ROOT = os.environ.get('FOGLAMP_ROOT')
SOUTH_SERVICE_NAME = "Sine #1"
SOUTH_DP_NAME="sinusoid"
SOUTH_ASSET_NAME = "{}_sinusoid_assets".format(time.strftime("%Y%m%d"))
NORTH_PLUGIN = "OMF"
NOTIF_SERVICE_NAME = "notification"
NOTIF_INSTANCE_NAME = "notify #1"
AF_HIERARCHY_LEVEL = "{0}_teststatslvl1/{0}_teststatslvl2/{0}_teststatslvl3".format(time.strftime("%Y%m%d"))

@pytest.fixture
def reset_foglamp(wait_time):
    try:
        subprocess.run(["cd {}/tests/system/python/scripts/package && ./reset"
                       .format(PROJECT_ROOT)], shell=True, check=True)
    except subprocess.CalledProcessError:
        assert False, "reset package script failed!"

    # Wait for foglamp server to start
    time.sleep(wait_time)

@pytest.fixture
def reset_eds(north_historian):
    if north_historian != "EdgeDataStore":
        print("Skipping EDS reset check as north_historian is not EdgeDataStore")
        return
    eds_reset_url = "/api/v1/Administration/Storage/Reset"
    con = http.client.HTTPConnection("localhost", 5590)
    con.request("POST", eds_reset_url, "")
    resp = con.getresponse()
    assert 204 == resp.status

@pytest.fixture
def check_eds_installed(north_historian):
    
    if north_historian != "EdgeDataStore":
        print("Skipping EDS installation check as north_historian is not EdgeDataStore")
        return

    dpkg_list = os.popen('dpkg --list osisoft.edgedatastore >/dev/null; echo $?')
    ls_output = dpkg_list.read()
    assert ls_output == "0\n", "EDS not installed. Please install it first!"
    eds_data_url = "/api/v1/diagnostics/productinformation"
    con = http.client.HTTPConnection("localhost", 5590)
    con.request("GET", eds_data_url)
    resp = con.getresponse()
    r = json.loads(resp.read().decode())
    assert len(r) != 0, "EDS not installed. Please install it first!"

@pytest.fixture
def start_south(add_south, foglamp_url):
    south_plugin = "sinusoid"
    config = {"assetName": {"value": SOUTH_ASSET_NAME}}
    # south_branch does not matter as these are archives.foglamp.org version install
    add_south(south_plugin, None, foglamp_url, service_name=SOUTH_SERVICE_NAME, installation_type='package', config=config)


def _verify_egress(read_data_from_pi_web_api, pi_host, pi_admin, pi_passwd, pi_db, wait_time, retries):
    retry_count = 0
    data_from_pi = None
    af_hierarchy_level_list = AF_HIERARCHY_LEVEL.split("/")

    while (data_from_pi is None or data_from_pi == []) and retry_count < retries:
        data_from_pi = read_data_from_pi_web_api(pi_host, pi_admin, pi_passwd, pi_db, af_hierarchy_level_list,
                                                 SOUTH_ASSET_NAME, '')
        retry_count += 1
        time.sleep(wait_time * 2)

    if data_from_pi is None or retry_count == retries:
        assert False, "Unable to read data from PI"

@pytest.fixture
def start_north(foglamp_url, north_historian, start_north_omf_as_a_service, 
                pi_host, pi_port, pi_admin, pi_passwd, enabled=True):
    
    if north_historian == "EdgeDataStore":
        data = {"name": "EDS #1",
                "plugin": NORTH_PLUGIN,
                "type": "north",
                "enabled": enabled,
                "config": {"PIServerEndpoint": {"value": "Edge Data Store"},
                        "NamingScheme": {"value": "Backward compatibility"}}
                }
        post_url = "/foglamp/service"
        utils.post_request(foglamp_url, post_url, data)
    else:
        start_north_omf_as_a_service(foglamp_url, pi_host, pi_port, pi_user=pi_admin, pi_pwd=pi_passwd, pi_use_legacy="false",
                                     service_name="OMF #1", default_af_location=AF_HIERARCHY_LEVEL)

@pytest.fixture
def start_notification(foglamp_url, add_service, add_notification_instance,wait_time, retries):
    
    # Install and Add Notification Service
    add_service(foglamp_url, "notification", None, retries, installation_type='package', service_name=NOTIF_SERVICE_NAME)
    
    # Wait and verify service created or not
    time.sleep(wait_time)
    verify_service_added(foglamp_url, NOTIF_SERVICE_NAME)
    
    # Add Notification Instance
    rule_config = {"auditCode": "CONAD,SCHAD"}
    delivery_config = {"enable": "true"}
    add_notification_instance(foglamp_url, "asset", None, rule_config=rule_config, delivery_config=delivery_config, 
                              rule_plugin="DataAvailability", installation_type='package', notification_type="retriggered",
                              notification_instance_name="test #1", retrigger_time=5)
    
    # Verify Notification Instance created or not
    notification_url = "/foglamp/notification"
    resp = utils.get_request(foglamp_url, notification_url)
    assert "test #1" in [s["name"] for s in resp["notifications"]]

def verify_service_added(foglamp_url, name):
    get_url = "/foglamp/service"
    result = utils.get_request(foglamp_url, get_url)
    assert len(result["services"])
    assert name in [s["name"] for s in result["services"]]

def verify_ping(foglamp_url, skip_verify_north_interface, wait_time, retries):
    get_url = "/foglamp/ping"
    ping_result = utils.get_request(foglamp_url, get_url)
    assert "dataRead" in ping_result
    assert "dataSent" in ping_result
    assert 0 < ping_result['dataRead'], "South data NOT seen in ping header"

    retry_count = 1
    sent = 0
    if not skip_verify_north_interface:
        while retries > retry_count:
            sent = ping_result["dataSent"]
            if sent >= 1:
                break
            else:
                time.sleep(wait_time)

            retry_count += 1
            ping_result = utils.get_request(foglamp_url, get_url)

        assert 1 <= sent, "Failed to send data to Edge Data Store"
    return ping_result

def verify_eds_data():
    eds_data_url = "/api/v1/tenants/default/namespaces/default/streams/1measurement_{}/Data/Last".format(SOUTH_ASSET_NAME)
    print (eds_data_url)
    con = http.client.HTTPConnection("localhost", 5590)
    con.request("GET", eds_data_url)
    resp = con.getresponse()
    r = json.loads(resp.read().decode())
    return r

class TestDataAvailabilityAuditBasedNotificationRuleOnIngress:
    def test_data_availability_multiple_audit(self, clean_setup_foglamp_packages, reset_foglamp, start_notification, 
                                              start_south, foglamp_url, skip_verify_north_interface, wait_time, retries):
        """ Test NTFSN triggered or not with CONAD, SCHAD.
            clean_setup_foglamp_packages: Fixture to remove and install latest foglamp packages
            reset_foglamp: Fixture to reset foglamp
            start_south: Fixtures to add and start south services
            start_notification: Fixture to add and start notification service with rule and delivery plugins
            Assertions:
                on endpoint GET /foglamp/audit
                on endpoint GET /foglamp/ping
                on endpoint GET /foglamp/category """
        time.sleep(wait_time)

        verify_ping(foglamp_url, True, wait_time, retries)

        get_url = "/foglamp/audit?source=NTFSN"
        resp1 = utils.get_request(foglamp_url, get_url)
        print (len(resp1['audit']))
        assert len(resp1['audit'])
        
        assert "test #1" in [s["details"]["name"] for s in resp1["audit"]]
        for audit_detail in resp1['audit']:
            if "test #1" == audit_detail['details']['name']:
                assert "NTFSN" == audit_detail['source'], "ERROR: NTFSN not triggered properly on CONAD or SCHAD"

    def test_data_availability_single_audit(self, foglamp_url, skip_verify_north_interface, wait_time, retries):
        """ Test NTFSN triggered or not with CONCH in sinusoid plugin.
                on endpoint GET /foglamp/audit
                on endpoint GET /foglamp/ping
                on endpoint GET /foglamp/category """
        get_url = "/foglamp/audit?source=NTFSN"
        resp1 = utils.get_request(foglamp_url, get_url)

        # Change the configuration of rule plugin
        put_url = "/foglamp/category/ruletest #1"
        data = {"auditCode": "CONCH"}
        utils.put_request(foglamp_url, urllib.parse.quote(put_url), data)
        
        # Change the configuration of sinusoid plugin
        put_url = "/foglamp/category/Sine #1Advanced"
        data = {"readingsPerSec": "10"}
        utils.put_request(foglamp_url, urllib.parse.quote(put_url), data)

        time.sleep(wait_time)
        get_url = "/foglamp/audit?source=NTFSN"
        resp2 = utils.get_request(foglamp_url, get_url)
        assert len(resp2['audit']) - len(resp1['audit']) == 1, "ERROR: NTFSN not triggered properly on CONCH"

    @pytest.mark.xfail(reason="FOGL-7712")
    def test_data_availability_all_audit(self, foglamp_url, add_south, skip_verify_north_interface, wait_time, retries):
        """ Test NTFSN triggered or not with all audit changes.
                on endpoint GET /foglamp/audit
                on endpoint GET /foglamp/ping
                on endpoint GET /foglamp/category """
        get_url = "/foglamp/audit?source=NTFSN"
        resp1 = utils.get_request(foglamp_url, get_url)

        # Change the configuration of rule plugin
        put_url = "/foglamp/category/ruletest #1"
        data = {"auditCode": "*"}
        utils.put_request(foglamp_url, urllib.parse.quote(put_url), data)
        
        # Add new service
        south_plugin = "sinusoid"
        config = {"assetName": {"value": "sine-test"}}
        # south_branch does not matter as these are archives.foglamp.org version install
        add_south(south_plugin, None, foglamp_url, service_name="sine-test", installation_type='package', config=config)

        time.sleep(wait_time)
        get_url = "/foglamp/audit?source=NTFSN"
        resp2 = utils.get_request(foglamp_url, get_url)
        assert len(resp2['audit']) > len(resp1['audit']), "ERROR: NTFSN not triggered properly with * audit code"

class TestDataAvailabilityAssetBasedNotificationRuleOnIngress:
    def test_data_availability_asset(self, foglamp_url, add_south, skip_verify_north_interface, wait_time, retries):
        """ Test NTFSN triggered or not with all audit changes.
                on endpoint GET /foglamp/audit
                on endpoint GET /foglamp/ping
                on endpoint GET /foglamp/category """
        get_url = "/foglamp/audit?source=NTFSN"
        resp1 = utils.get_request(foglamp_url, get_url)

        # Change the configuration of rule plugin
        put_url = "/foglamp/category/ruletest #1"
        data = {"auditCode": "", "assetCode": SOUTH_ASSET_NAME}
        utils.put_request(foglamp_url, urllib.parse.quote(put_url), data)

        time.sleep(wait_time)
        get_url = "/foglamp/audit?source=NTFSN"
        resp2 = utils.get_request(foglamp_url, get_url)
        assert len(resp2['audit']) > len(resp1['audit']), "ERROR: NTFSN not triggered properly with asset code"

class TestDataAvailabilityBasedNotificationRuleOnEgress:
    def test_data_availability_north(self, check_eds_installed, reset_foglamp, start_notification, reset_eds, 
                                     start_north, foglamp_url, wait_time, skip_verify_north_interface, add_south, 
                                     retries, north_historian, read_data_from_pi_web_api, pi_host, pi_admin, pi_passwd, 
                                     pi_db):
        """ Test NTFSN triggered or not with configuration change in north EDS plugin.
            start_north: Fixtures to add and start south services
            Assertions:
                on endpoint GET /foglamp/audit
                on endpoint GET /foglamp/ping """
        
        # Change the configuration of rule plugin
        put_url = "/foglamp/category/ruletest #1"
        data = {"auditCode": "", "assetCode": SOUTH_ASSET_NAME}
        utils.put_request(foglamp_url, urllib.parse.quote(put_url), data)
        
        # Add new service
        south_plugin = "sinusoid"
        config = {"assetName": {"value": SOUTH_ASSET_NAME}}
        # south_branch does not matter as these are archives.foglamp.org version install
        add_south(south_plugin, None, foglamp_url, service_name="sine-test", installation_type='package', config=config)

        get_url = "/foglamp/audit?source=NTFSN"
        resp1 = utils.get_request(foglamp_url, get_url)
        
        time.sleep(wait_time)
        get_url = "/foglamp/audit?source=NTFSN"
        resp2 = utils.get_request(foglamp_url, get_url)
        assert len(resp2['audit']) > len(resp1['audit']), "ERROR: NTFSN not triggered properly with asset code"
        
        time.sleep(wait_time)
        verify_ping(foglamp_url, skip_verify_north_interface, wait_time, retries)
        if north_historian == "EdgeDataStore":
            r = verify_eds_data()
            assert SOUTH_DP_NAME in r, "Data found in EDS!"
            ts = r.get("Time")
            assert ts.find(datetime.now().strftime("%Y-%m-%d")) != -1, "Latest data not found in EDS!"
        else:
            _verify_egress(read_data_from_pi_web_api, pi_host, pi_admin, pi_passwd, pi_db, wait_time, retries)
